#include "pPanels.h"
#include<iostream>
#include <Eigen/Dense>
#include <cmath>
#include "math.h"

using namespace sdpm;

Panels::Panels(size_t _nPanChord)
{
    nPanChord = _nPanChord;
    nPanSpan = 2;
    beta = 1;
}

void Panels::CreateWing()
{
    Eigen::RowVectorXd xp(2*nPanChord+1); // Non-linear chordwise distribution
    Eigen::RowVectorXd xc(2*nPanChord+1); // Control point chordwise distribution

    Eigen::RowVectorXd dumm = Eigen::VectorXd::LinSpaced(2*nPanChord+1, 0, M_PI);
    for (auto i=0; i<xp.size(); ++i)
    {
        xp(i) = 1-std::sin(dumm(i));
    }

    for (auto i=0; i<xp.size(); ++i)
    {
        xc(i) = (xp(i+1) + xp(i))/2;
    }
    
    /* Non-linear spanwise distribution */

    Eigen::RowVectorXd yp(nPanSpan+1);
    Eigen::RowVectorXd yc(nPanSpan);
    Eigen::RowVectorXd theta = Eigen::RowVectorXd::LinSpaced(nPanSpan+1, M_PI, 0);
    for (auto i=0; i<theta.size(); ++i)
    {
        yp(i) = std::cos(theta(i));
    }
    
    for (auto i=0; i<yc.size(); ++i)
    {
        yc(i) = (yp(i+1) + yp(i)) / 2;
    } 

    /* Spanwise chord variation */
    Eigen::RowVectorXd cvar;
    cvar = (lambda-1)*c0*yp.cwiseAbs();
    cvar += c0 * Eigen::MatrixXd::Ones(1, cvar.size());
    Eigen::RowVectorXd epsilonvar;
    epsilonvar=epsilon*yp.cwiseAbs();

    /*  Spanwise quarter chord distance due to twist */
    Eigen::RowVectorXd xsweep;
    xsweep = std::tan(LambdaC4)*yp.cwiseAbs()*b/2;
    /* Spanwise vertical distance due to dihedral */
    Eigen::RowVectorXd zdih;
    zdih = std::tan(dihedral)*yp.cwiseAbs()*b/2;
    

    /* Calculate airfoil shape at each panel gridpoint */

    Eigen::MatrixXd zp = Eigen::MatrixXd::Zero(1,2*nPanChord+1);

    double t = .12; // Airfoil maximum thickness

    /* Upper side */

    for (auto i=0; i<zp.size()/2; ++i)
    {
        zp(i) = -(t/0.2)*(0.2969*std::sqrt(xp(i)) - .126 * xp(i) - .3516* xp(i)*xp(i) + .2843*xp(i)*xp(i)*xp(i) - .1036*xp(i)*xp(i)*xp(i)*xp(i));
    }

    /* Lower side */

    for (auto i=zp.size()/2; i<zp.size(); ++i)
    {
        zp(i) = (t/0.2)*(0.2969*std::sqrt(xp(i)) - .126 * xp(i) - .3516* xp(i)*xp(i) + .2843*xp(i)*xp(i)*xp(i) - .1036*xp(i)*xp(i)*xp(i)*xp(i));
    }

    // iko=find(xp <= maxpos);
    // zp(iko)=zp(iko)+maxcamb/maxpos^2*(2*maxpos*xp(iko)-xp(iko).^2);
    // iko=find(xp > maxpos);
    // zp(iko)=zp(iko)+maxcamb/(1-maxpos)^2*((1-2*maxpos)+2*maxpos*xp(iko)-xp(iko).^2);

    /* Assemble bound panel corner point matrices */
    /* Eigen::Array * Eigen::Array is a mult element wise while Eigen::Matrix * Eigen::Matrix is a matrix product -> cwiseProduct */
    Xp = xp.transpose().replicate(1,nPanSpan+1).cwiseProduct(cvar.replicate(2*nPanChord+1,1)) - cvar.replicate(2*nPanChord+1,1)*ax;
    Yp = yp.replicate(2*nPanChord+1,1) * b/2;
    Zp = zp.transpose().replicate(1,nPanSpan+1).cwiseProduct(cvar.replicate(2*nPanChord+1,1));

    Surface = (c0+c0*lambda)/2 * b;



    // % Apply twist
    // for i=1:n+1
    //     Ry=(cos(epsilonvar(i)) sin(epsilonvar(i));-sin(epsilonvar(i)) cos(epsilonvar(i)));
    //     dummy=Ry*(Xp(:,i)';Zp(:,i)');
    //     Xp(:,i)=dummy(1,:)';
    //     Zp(:,i)=dummy(2,:)';
    // end

    // % Apply dihedral
    // Zp=Zp+repmat(zdih,2*m+1,1);
    // % Apply sweep
    // Xp=Xp+repmat(xsweep,2*m+1,1);

    Xc.setZero(2*nPanChord, nPanSpan);
    Yc.setZero(2*nPanChord, nPanSpan);
    Zc.setZero(2*nPanChord, nPanSpan);
    nx.setZero(2*nPanChord, nPanSpan);
    ny.setZero(2*nPanChord, nPanSpan);
    nz.setZero(2*nPanChord, nPanSpan);
    Xc0.setZero(2*nPanChord, nPanSpan);
    Yc0.setZero(2*nPanChord, nPanSpan);
    Zc0.setZero(2*nPanChord, nPanSpan);
    nx0.setZero(2*nPanChord, nPanSpan);
    ny0.setZero(2*nPanChord, nPanSpan);
    nz0.setZero(2*nPanChord, nPanSpan);
    tauxx.setZero(2*nPanChord, nPanSpan);
    tauxy.setZero(2*nPanChord, nPanSpan);
    tauxz.setZero(2*nPanChord, nPanSpan);
    tauyx.setZero(2*nPanChord, nPanSpan);
    tauyy.setZero(2*nPanChord, nPanSpan);
    tauyz.setZero(2*nPanChord, nPanSpan);
    s.setZero(2*nPanChord, nPanSpan);
    s0.setZero(2*nPanChord, nPanSpan);
    cpln.setZero(2*nPanChord, nPanSpan);

    for (size_t i=0; i<2*nPanChord; ++i)
    {
        for (size_t j=0; j<nPanSpan; ++j)
        {
            /* Calculate control points */
            Xc0(i,j) = (Xp(i,j) + Xp(i,j+1) + Xp(i+1,j+1) + Xp(i+1,j))/4;
            Yc0(i,j) = (Yp(i,j) + Yp(i,j+1) + Yp(i+1,j+1) + Yp(i+1,j))/4;
            Zc0(i,j) = (Zp(i,j) + Zp(i,j+1) + Zp(i+1,j+1) + Zp(i+1,j))/4;

            /* Calculate the normal vectors */
            /*(N,s0(i,j),cpln(i,j))=normalarea((Xp(i,j) Xp(i,j+1) Xp(i+1,j+1) Xp(i+1,j)), ...
            (Yp(i,j) Yp(i,j+1) Yp(i+1,j+1) Yp(i+1,j)), ...
            (Zp(i,j) Zp(i,j+1) Zp(i+1,j+1) Zp(i+1,j))); */

            /* Normal vector (cross product) */
            Eigen::RowVector3d normal1 = Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i,j+1)-Xp(i+1,j), Yp(i,j+1)-Yp(i+1,j), Zp(i,j+1)-Zp(i+1,j)));
            Eigen::RowVectorXd mag1 = (normal1*normal1.transpose());
            mag1(0) = std::sqrt(mag1(0));
            normal1 /= mag1(0); // mag1 is of size 1

            /* Calculate a second normal from points 1,2,3 only */
            Eigen::RowVector3d normal2= Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i,j+1)-Xp(i,j), Yp(i,j+1)-Yp(i,j), Zp(i,j+1)-Zp(i,j)));
            Eigen::RowVectorXd mag2=(normal2*normal2.transpose());
            mag2(0) = std::sqrt(mag2(0));
            normal2=normal2/mag2(0);

            /* Test if the two normals are colinear */
            cpln(i,j) = normal1.cross(normal2)(0);
            Eigen::RowVectorXd dum1 = cpln*cpln.transpose();
            cpln(i,j)=sqrt(dum1(0));

            /* Calculate a third normal from points 1,3,4 only */
            Eigen::RowVector3d normal3 = Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i+1,j)-Xp(i,j), Yp(i+1,j)-Yp(i,j), Zp(i+1,j)-Zp(i,j)));
            Eigen::RowVectorXd mag3=(normal3*normal3.transpose());
            mag3(0) = std::sqrt(mag3(0));
            /* Panel area */
            s0(i,j) = mag2(0)/2+mag3(0)/2;

            nx0(i,j)  = normal1(0);
            ny0(i,j)  = normal1(1);
            nz0(i,j)  = normal1(2);

        }
    }

    /* Scale for compressibility */
    /*Xp0=Xp;
    Yp0=Yp;
    Zp0=Zp;
    Xp=Xp0/beta; // (beta = sqrt(1-M*M))
    Yp=Yp0;
    Zp=Zp0;*/

    for (size_t i=0 ; i<2*nPanChord; ++i)
    {
        for (size_t j=0; j<nPanSpan; ++j)
        {
            /* Calculate control points */
            Xc(i,j)=(Xp(i,j) + Xp(i,j+1) + Xp(i+1,j+1) + Xp(i+1,j))/4;
            Yc(i,j)=(Yp(i,j) + Yp(i,j+1) + Yp(i+1,j+1) + Yp(i+1,j))/4;
            Zc(i,j)=(Zp(i,j) + Zp(i,j+1) + Zp(i+1,j+1) + Zp(i+1,j))/4;
            /* Calculate the normal vectors */
            /* (N,s(i,j),cpln(i,j))=normalarea((Xp(i,j) Xp(i,j+1) Xp(i+1,j+1) Xp(i+1,j)), ...
                (Yp(i,j) Yp(i,j+1) Yp(i+1,j+1) Yp(i+1,j)), ...
                (Zp(i,j) Zp(i,j+1) Zp(i+1,j+1) Zp(i+1,j))); */

            /* Normal vector (cross product) */
            Eigen::RowVector3d normal1 = Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i,j+1)-Xp(i+1,j), Yp(i,j+1)-Yp(i+1,j), Zp(i,j+1)-Zp(i+1,j)));
            Eigen::RowVectorXd mag1 = (normal1*normal1.transpose());
            mag1(0) = std::sqrt(mag1(0));
            normal1 /= mag1(0); // mag1 is of size 1

            /* Calculate a second normal from points 1,2,3 only */
            Eigen::RowVector3d normal2= Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i,j+1)-Xp(i,j), Yp(i,j+1)-Yp(i,j), Zp(i,j+1)-Zp(i,j)));
            Eigen::RowVectorXd mag2=(normal2*normal2.transpose());
            mag2(0) = std::sqrt(mag2(0));
            normal2=normal2/mag2(0);

            /* Test if the two normals are colinear */
            cpln(i,j) = normal1.cross(normal2)(0);
            Eigen::RowVectorXd dum1 = cpln*cpln.transpose();
            cpln(i,j)=sqrt(dum1(0));

            /* Calculate a third normal from points 1,3,4 only */
            Eigen::RowVector3d normal3 = Eigen::RowVector3d(Xp(i+1,j+1)-Xp(i,j), Yp(i+1,j+1)-Yp(i,j), Zp(i+1,j+1)-Zp(i,j)).cross(Eigen::RowVector3d(Xp(i+1,j)-Xp(i,j), Yp(i+1,j)-Yp(i,j), Zp(i+1,j)-Zp(i,j)));
            Eigen::RowVectorXd mag3=(normal3*normal3.transpose());
            mag3(0) = std::sqrt(mag3(0));
            /* Panel area */
            s(i,j) = mag2(0)/2+mag3(0)/2;

            nx(i,j)  = normal1(0);
            ny(i,j)  = normal1(1);
            nz(i,j)  = normal1(2);
            
            /* Calculate tangential vector in x direction */
            double a=1/sqrt(1+(normal1(0)/normal1(2))*(normal1(0)/normal1(2)));
            double c=-normal1(0)/normal1(2)*a;
            tauxx(i,j)=a;
            tauxy(i,j)=0;
            tauxz(i,j)=c;
            
            /* Calculate tangential vector in y direction */
            Eigen::RowVector3d dummy = normal1.cross(Eigen::RowVector3d(a, 0, c));
            tauyx(i,j)=dummy(0);
            tauyy(i,j)=dummy(1);
            tauyz(i,j)=dummy(2);
        } 
    }

}

void Panels::WakeControlPoints(unsigned long it){
    for (size_t i=0; i<it; ++i){
            for (size_t j=0; j<nPanSpan; ++j)
            {

                /* Calculate control points */
                Xcw(i,j) = (Xw(i,j) + Xw(i,j+1) + Xw(i+1,j+1) + Xw(i+1,j))/4;
                Ycw(i,j) = (Yw(i,j) + Yw(i,j+1) + Yw(i+1,j+1) + Yw(i+1,j))/4;
                Zcw(i,j) = (Zw(i,j) + Zw(i,j+1) + Zw(i+1,j+1) + Zw(i+1,j))/4;
                /* Calculate the normal vectors */
                /* N=normalarea((Xw(i,j) Xw(i,j+1) Xw(i+1,j+1) Xw(i+1,j)), ...
                    (Yw(i,j) Yw(i,j+1) Yw(i+1,j+1) Yw(i+1,j)), ...
                    (Zw(i,j) Zw(i,j+1) Zw(i+1,j+1) Zw(i+1,j))); */

                /* Normal vector (cross product) */
                Eigen::RowVector3d normal1 = Eigen::RowVector3d(Xw(i+1,j+1)-Xw(i,j), Yw(i+1,j+1)-Yw(i,j), Zw(i+1,j+1)-Zw(i,j)).cross(Eigen::RowVector3d(Xw(i,j+1)-Xw(i+1,j), Yw(i,j+1)-Yw(i+1,j), Zw(i,j+1)-Zw(i+1,j)));
                Eigen::RowVectorXd mag1 = (normal1*normal1.transpose());
                mag1(0) = std::sqrt(mag1(0));
                normal1 /= mag1(0); // mag1 is of size 1

                nxw(i,j)  = normal1(0);
                nyw(i,j)  = normal1(1);
                nzw(i,j)  = normal1(2);
            
                /* Calculate tangential vector in x direction */
                double a= 1/sqrt(1+(normal1(0)/normal1(2))*(normal1(0)/normal1(2)));
                double c=-normal1(0)/normal1(2)*a;
                tauxxw(i,j)=a;
                tauxyw(i,j)=0;
                tauxzw(i,j)=c;
            
                /* Calculate tangential vector in y direction */
                Eigen::Vector3d dummy = normal1.cross(Eigen::Vector3d(a, 0, c));
                tauyxw(i,j)=dummy(0);
                tauyyw(i,j)=dummy(1);
                tauyzw(i,j)=dummy(2);
            }
        }
}
        

void Panels::ComputeInfluenceCoeffs_wing()
{
    /* Sources */
    AN.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Au.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Av.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Aw.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    /* Doublets */
    BN.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Bu.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Bv.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    Bw.setZero(2*nPanChord*nPanSpan,2*nPanChord*nPanSpan);
    /* Call mex function to calculate influence coefficients of wing on wing */

    Eigen::RowVectorXd x, y, z;
    x.setZero(5);
    y.setZero(5);
    z.setZero(5);

    Eigen::RowVector3d normal;
    Eigen::RowVector3d taux;
    Eigen::RowVector3d tauy;

    Eigen::RowVector2d Phisd;
    Eigen::RowVectorXd uvw;
    uvw.setZero(6);

    double x0, y0, z0, xcp, ycp, zcp;

    size_t ic,jc,i,j;
    
    size_t m = Xc.rows();
    size_t n = Xc.cols();
    //size_t mp1=m+1;
    //size_t np1=n+1;
    //size_t mn=m*n;
    
    /* Create Influence Matrix */
    
    for (ic=0; ic < m; ic++) {
        for (jc=0; jc < n; jc++) {
            for (i=0; i < m; i++) {
                for (j=0; j < n; j++) {
                    x0=Xc(ic,jc);
                    y0=Yc(ic,jc);
                    z0=Zc(ic,jc);
                    xcp=Xc(i,j);
                    ycp=Yc(i,j);
                    zcp=Zc(i,j);
                    x(0)=Xp(i,j);
                    x(1)=Xp(i,(j+1));
                    x(2)=Xp(i+1,(j+1));
                    x(3)=Xp(i+1,j);
                    x(4)=Xp(i,j);
                    y(0)=Yp(i,j);
                    y(1)=Yp(i,(j+1));
                    y(2)=Yp(i+1,(j+1));
                    y(3)=Yp(i+1,j);
                    y(4)=Yp(i,j);
                    z(0)=Zp(i,j);
                    z(1)=Zp(i,(j+1));
                    z(2)=Zp(i+1,(j+1));
                    z(3)=Zp(i+1,j);
                    z(4)=Zp(i,j);
                    normal(0)=nx(i,j);
                    normal(1)=ny(i,j);
                    normal(2)=nz(i,j);
                    taux(0)=tauxx(i,j);
                    taux(1)=tauxy(i,j);
                    taux(2)=tauxz(i,j);
                    tauy(0)=tauyx(i,j);
                    tauy(1)=tauyy(i,j);
                    tauy(2)=tauyz(i,j);
                    
                    sdpanel_wing(Phisd,uvw,x0,y0,z0,x,y,z,xcp,ycp,zcp,normal,taux,tauy);
                    AN(ic*n+jc,i*n+j)=Phisd(0); 
                    Bu(ic*n+jc,i*n+j)=-uvw(3); 
                    Bv(ic*n+jc,i*n+j)=-uvw(4); 
                    Bw(ic*n+jc,i*n+j)=-uvw(5);                     
                    if ((ic*n+jc) == (i*n+j)) {
                        Au(ic*n+jc,i*n+j)=normal(0)/2.0;
                        Av(ic*n+jc,i*n+j)=normal(1)/2.0;
                        Aw(ic*n+jc,i*n+j)=normal(2)/2.0;
                        BN(ic*n+jc,i*n+j)=-0.5;
                    }else{
                        Au(ic*n+jc,i*n+j)=uvw(0);
                        Av(ic*n+jc,i*n+j)=uvw(1);
                        Aw(ic*n+jc,i*n+j)=uvw(2);
                        BN(ic*n+jc,i*n+j)=Phisd(1);
                    }

                }
            }
        }
    }
}

void Panels::sdpanel_wing(Eigen::RowVector2d &Phisd, Eigen::RowVectorXd &uvw, double x0, double y0, double z0, Eigen::RowVectorXd x, Eigen::RowVectorXd y, Eigen::RowVectorXd z, double xcp, double ycp, double zcp,
        Eigen::RowVector3d normal, Eigen::RowVector3d taux, Eigen::RowVector3d tauy)
{

    Eigen::RowVectorXd xl, yl, zl, r, e, h, mij, dij, rij, wij;
    xl.setZero(5);
    yl.setZero(5);
    zl.setZero(5);
    r.setZero(5);
    e.setZero(5);
    h.setZero(5);

    mij.setZero(4);
    dij.setZero(4);
    rij.setZero(4);
    wij.setZero(4);

    double x0l,y0l,z0l;
    double pi,Phi1,Phi2,logterm,logdenom,atanterm,fraction,fracdenom,xx,yy,us,vs,ws,ud,vd,wd;
    int i;
    
    pi=3.14159265358979;

    for (i=0; i < 5; i++) { 
        xl(i)=(x(i)-xcp)*taux(0)+(y(i)-ycp)*taux(1)+(z(i)-zcp)*taux(2);
        yl(i)=(x(i)-xcp)*tauy(0)+(y(i)-ycp)*tauy(1)+(z(i)-zcp)*tauy(2);
        zl(i)=(x(i)-xcp)*normal(0)+(y(i)-ycp)*normal(1)+(z(i)-zcp)*normal(2);
   }
    x0l=(x0-xcp)*taux(0)+(y0-ycp)*taux(1)+(z0-zcp)*taux(2);
    y0l=(x0-xcp)*tauy(0)+(y0-ycp)*tauy(1)+(z0-zcp)*tauy(2);
    z0l=(x0-xcp)*normal(0)+(y0-ycp)*normal(1)+(z0-zcp)*normal(2);
    
    for (i=0; i < 4; i++) { 
        r(i)=sqrt((x0l-xl(i))*(x0l-xl(i))+(y0l-yl(i))*(y0l-yl(i))+z0l*z0l);
        rij(i)=(x0l-xl(i))*(x0l-xl(i+1))+(y0l-yl(i))*(y0l-yl(i+1))+z0l*z0l;
        wij(i)=(x0l-xl(i+1))*(y0l-yl(i))-(x0l-xl(i))*(y0l-yl(i+1));
        e(i)=(x0l-xl(i))*(x0l-xl(i))+z0l*z0l;
        h(i)=(x0l-xl(i))*(y0l-yl(i));
        mij(i)=(yl(i+1)-yl(i))/(xl(i+1)-xl(i));
        dij(i)=sqrt((xl(i+1)-xl(i))*(xl(i+1)-xl(i))+(yl(i+1)-yl(i))*(yl(i+1)-yl(i)));
    }
    r(4)=r(0);
    e(4)=e(0);
    h(4)=h(0);
    
    Phi1=0.;
    Phi2=0.;
    us=0.;
    vs=0.;
    ud=0.;
    vd=0.;
    wd=0.;
    for (i=0; i < 4; i++) {
        logterm=0.0;
        logdenom=(r(i)+r(i+1)-dij(i));
        if (fabs(logdenom) > 1e-10){
            logterm=log((r(i)+r(i+1)+dij(i))/logdenom);
        }
        us=us-(yl(i+1)-yl(i))*logterm/dij(i);
        vs=vs-(xl(i)-xl(i+1))*logterm/dij(i);
        Phi1+=((x0l-xl(i))*(yl(i+1)-yl(i))-(y0l-yl(i))*(xl(i+1)-xl(i)))/dij(i)*logterm;
        atanterm=0.0;
        if ((fabs(xl(i+1)-xl(i)) > 1e-10) && (fabs(z0l) > 1e-10)){ //mij(i) is not infinite
            xx=(mij(i)*e(i)-h(i))/(z0l*r(i));
            yy=(mij(i)*e(i+1)-h(i+1))/(z0l*r(i+1));
            atanterm=atan2(xx-yy,1+xx*yy);
        }
        fraction=0.0;
        fracdenom=r(i)*r(i+1)*(r(i)*r(i+1)+rij(i));
        if (fabs(fracdenom) > 1e-10){
            fraction=(r(i)+r(i+1))/fracdenom;
        }
        Phi2+=atanterm;
        ud+=z0l*(yl(i)-yl(i+1))*fraction;
        vd+=z0l*(xl(i+1)-xl(i))*fraction;
        wd+=wij(i)*fraction;
    }
    // Source potential
    Phisd(0)=-1/4.0/pi*(Phi1+fabs(z0l)*Phi2);
//    Phisd(0)=-1/4.0/pi*(Phi1-z0l*Phi2);
    // Source velocities
    us=us/4.0/pi;
    vs=vs/4.0/pi;
    ws=Phi2/4.0/pi;
    // Doublet potential
    Phisd(1)=ws;
    // Doublet velocities
    ud=ud/4.0/pi;
    vd=vd/4.0/pi;
    wd=wd/4.0/pi;
    
    // Transform velocities to global coordinates
    uvw(0)=taux(0)*us+tauy(0)*vs+normal(0)*ws;
    uvw(1)=taux(1)*us+tauy(1)*vs+normal(1)*ws;
    uvw(2)=taux(2)*us+tauy(2)*vs+normal(2)*ws;
    uvw(3)=taux(0)*ud+tauy(0)*vd+normal(0)*wd;
    uvw(4)=taux(1)*ud+tauy(1)*vd+normal(1)*wd;
    uvw(5)=taux(2)*ud+tauy(2)*vd+normal(2)*wd;
}

void Panels::ComputeInfluenceCoeffs_wake(unsigned long it)
{

    CN.setZero(2*nPanChord*nPanSpan, nPanSpan*it);
    Cu.setZero(2*nPanChord*nPanSpan, nPanSpan*it);
    Cv.setZero(2*nPanChord*nPanSpan, nPanSpan*it);
    Cw.setZero(2*nPanChord*nPanSpan, nPanSpan*it);

    Eigen::RowVectorXd x, y, z;
    x.setZero(5);
    y.setZero(5);
    z.setZero(5);

    Eigen::RowVector3d normal;
    Eigen::RowVector3d taux;
    Eigen::RowVector3d tauy;

    Eigen::RowVectorXd Phisd;
    Phisd.setZero(1);
    Eigen::RowVectorXd uvw;
    uvw.setZero(3);

    double x0, y0, z0, xcp, ycp, zcp;
    
    size_t ic,jc,i,j;
    
    size_t m = Xc.rows();
    size_t n = Xc.cols();
    //size_t mp1=m+1;
    //size_t np1=n+1;
    //size_t mn=m*n;

    size_t mwp1 = it+1;
    size_t mw = mwp1-1;
    
    /* Create Influence Matrix */
    
    for (ic=0; ic < m; ic++) {
        for (jc=0; jc < n; jc++) {
            for (i=0; i < mw; i++) {
                for (j=0; j < n; j++) {
                    x0=Xc(ic,jc);
                    y0=Yc(ic,jc);
                    z0=Zc(ic,jc);
                    xcp=Xcw(i,j);
                    ycp=Ycw(i,j);
                    zcp=Zcw(i,j);
                    x(0)=Xw(i,j);
                    x(1)=Xw(i,(j+1));
                    x(2)=Xw(i+1,(j+1));
                    x(3)=Xw(i+1,j);
                    x(4)=Xw(i,j);
                    y(0)=Yw(i,j);
                    y(1)=Yw(i,(j+1));
                    y(2)=Yw(i+1,(j+1));
                    y(3)=Yw(i+1,j);
                    y(4)=Yw(i,j);
                    z(0)=Zw(i,j);
                    z(1)=Zw(i,(j+1));
                    z(2)=Zw(i+1,(j+1));
                    z(3)=Zw(i+1,j);
                    z(4)=Zw(i,j);
                    normal(0)=nxw(i,j);
                    normal(1)=nyw(i,j);
                    normal(2)=nzw(i,j);
                    taux(0)=tauxxw(i,j);
                    taux(1)=tauxyw(i,j);
                    taux(2)=tauxzw(i,j);
                    tauy(0)=tauyxw(i,j);
                    tauy(1)=tauyyw(i,j);
                    tauy(2)=tauyzw(i,j);
                    
                    sdpanel_wake(Phisd,uvw,x0,y0,z0,x,y,z,xcp,ycp,zcp,normal,taux,tauy);
                    CN(ic*n+jc, i*n+j)=Phisd(0);
                    Cu(ic*n+jc, i*n+j)=-uvw(0);
                    Cv(ic*n+jc, i*n+j)=-uvw(1);
                    Cw(ic*n+jc, i*n+j)=-uvw(2);
                }
            }
        }
    }
}

void Panels::sdpanel_wake(Eigen::RowVectorXd &Phisd, Eigen::RowVectorXd &uvw,double x0, double y0, double z0, Eigen::RowVectorXd x, Eigen::RowVectorXd y, Eigen::RowVectorXd z, double xcp, double ycp, double zcp,
        Eigen::RowVector3d normal, Eigen::RowVector3d taux, Eigen::RowVector3d tauy)
{

    Eigen::RowVectorXd xl, yl, zl, r, e, h, mij, dij, rij, wij;
    xl.setZero(5);
    yl.setZero(5);
    zl.setZero(5);
    r.setZero(5);
    e.setZero(5);
    h.setZero(5);

    mij.setZero(4);
    dij.setZero(4);
    rij.setZero(4);
    wij.setZero(4);

    double x0l,y0l,z0l;
    double pi,Phi2,atanterm,xx,yy,fraction,fracdenom,ud,vd,wd;
    int i;
    
    pi=3.14159265358979;
    
    for (i=0; i < 5; i++) { 
        xl(i)=(x(i)-xcp)*taux(0)+(y(i)-ycp)*taux(1)+(z(i)-zcp)*taux(2);
        yl(i)=(x(i)-xcp)*tauy(0)+(y(i)-ycp)*tauy(1)+(z(i)-zcp)*tauy(2);
        zl(i)=(x(i)-xcp)*normal(0)+(y(i)-ycp)*normal(1)+(z(i)-zcp)*normal(2);
   }
    x0l=(x0-xcp)*taux(0)+(y0-ycp)*taux(1)+(z0-zcp)*taux(2);
    y0l=(x0-xcp)*tauy(0)+(y0-ycp)*tauy(1)+(z0-zcp)*tauy(2);
    z0l=(x0-xcp)*normal(0)+(y0-ycp)*normal(1)+(z0-zcp)*normal(2);
    
    for (i=0; i < 4; i++) { 
        r(i)=sqrt((x0l-xl(i))*(x0l-xl(i))+(y0l-yl(i))*(y0l-yl(i))+z0l*z0l);
        rij(i)=(x0l-xl(i))*(x0l-xl(i+1))+(y0l-yl(i))*(y0l-yl(i+1))+z0l*z0l;
        wij(i)=(x0l-xl(i+1))*(y0l-yl(i))-(x0l-xl(i))*(y0l-yl(i+1));
        e(i)=(x0l-xl(i))*(x0l-xl(i))+z0l*z0l;
        h(i)=(x0l-xl(i))*(y0l-yl(i));
        mij(i)=(yl(i+1)-yl(i))/(xl(i+1)-xl(i));
        dij(i)=sqrt((xl(i+1)-xl(i))*(xl(i+1)-xl(i))+(yl(i+1)-yl(i))*(yl(i+1)-yl(i)));
    }
    r(4)=r(0);
    e(4)=e(0);
    h(4)=h(0);
    
    Phi2=0;
    ud=0.0;
    vd=0.0;
    wd=0.0;
    for (i=0; i < 4; i++) {
        atanterm=0;
        if ((xl(i+1)-xl(i) != 0) && (z0l != 0)){ //mij(i) is not infinite
            xx=(mij(i)*e(i)-h(i))/(z0l*r(i));
            yy=(mij(i)*e(i+1)-h(i+1))/(z0l*r(i+1));
            atanterm=atan2(xx-yy,1+xx*yy);
        }
        fraction=0.0;
        fracdenom=r(i)*r(i+1)*(r(i)*r(i+1)+rij(i));
        if (fabs(fracdenom) > 1e-10){
            fraction=(r(i)+r(i+1))/fracdenom;
        }
        Phi2+=atanterm;
        ud+=z0l*(yl(i)-yl(i+1))*fraction;
        vd+=z0l*(xl(i+1)-xl(i))*fraction;
        wd+=wij(i)*fraction;
    }
    // Doublet potential
    Phisd(0)=Phi2/4.0/pi;
    // Doublet velocities
    ud=ud/4.0/pi;
    vd=vd/4.0/pi;
    wd=wd/4.0/pi;

    // Transform velocities to global coordinates
    uvw(0)=taux(0)*ud+tauy(0)*vd+normal(0)*wd;
    uvw(1)=taux(1)*ud+tauy(1)*vd+normal(1)*wd;
    uvw(2)=taux(2)*ud+tauy(2)*vd+normal(2)*wd;
}



void Panels::SetKutta()
{
    /* Set up Kutta condition */
    Eigen::MatrixXd Kutta_dummy;
    Kutta_dummy.setZero(nPanSpan,2*nPanChord*nPanSpan);
    for (size_t i=0; i<nPanSpan; ++i)
    {
        Kutta_dummy(i,i)=1;
        Kutta_dummy(i,(2*nPanChord-1)*nPanSpan+i)=-1;
    }
    Kutta.setZero(Kutta_dummy.rows(), Kutta_dummy.cols()+(Eigen::MatrixXd::Identity(nPanSpan, nPanSpan).cols()));
    Kutta << Kutta_dummy, Eigen::MatrixXd::Identity(nPanSpan, nPanSpan);
}