from audioop import lin2adpcm
import numpy as np

class RbfInterpolator:

    def __init__(self) -> None:
        pass

    def RBFInterp(self, xs, ys, x, type, R):
        """"
        xs : numpy.ndarray Nx3, base mesh
        ys : numpy.ndarray Nx1, solution (on xs)
        x : numpy.ndarray Mx3, target mesh
        type : string
        R : double Radius"""
        if len(ys) != len(xs[:,0]):
            raise RuntimeError ("vector size does not match mesh", len(ys), "expected:",len(xs[:,0]))

        fPar = self.RBFparam(xs, ys, type, R)

        #if ~isempty(fPar)
        # Evaluate interpolation function
        y = self.RBFeval(xs, x, fPar, type, R);
        #else
        #    y = [];
        return y

    def RBFParam(self, xs, ys, RBFtype, R):

        if len(xs[:, 0]) == len(ys[:,0]):
            
            Ns = len(xs[:,0])
            Ms = len(ys)
            dim = len(xs[0,:])

            ncol = 1 + dim
            r = np.zeros(Ns)

            for i in range(Ns):
                for j in range(i,Ns):
                    r[i, j] = np.norm( xs[i, :] - xs[j, :])
                    r[j, i] = r[i, j]

            P = [np.ones((Ns, 1)), xs]
            M = self.radialFunction(r, RBFtype, R)
            
            A = [[M, P], [np.transpose(P), np.zeros(ncol)]]
            b = [[ys], [np.zeros(ncol, Ms)]]

            fPar = np.linalg.solve(A,b)
            
        else:
            
            fPar = []
            M = []
        return fPar

    def radialFunction(self, r, RBFtype, R):

        r = r/R;

        phi = np.zeros(len(r))

        if RBFtype == 'R1':
            phi = r
        elif RBFtype == 'R3':
            phi = r**3
        elif RBFtype == 'TPS2':
            phi[r>0] = r[r>0]**2*np.log(r[r>0])
        elif RBFtype == 'Q':
            phi = 1 + r**2
        elif RBFtype == 'MQ':
            phi = np.sqrt(1 + r**2)
        elif RBFtype == 'IMQ':
            phi = 1/np.sqrt(1 + r**2)
        elif RBFtype == 'IQ':
            phi = 1/(1 + r**2)
        elif RBFtype == 'GS':
            phi = np.exp(-r**2)
        elif RBFtype == 'CP_C0':
            I = (r < 1);
            phi[r<1] = (1 - r[r<1])**2
        elif RBFtype == 'CP_C2':
            I = (r < 1)
            phi[r<1] = (1 - r[r<1])**4*(4*r[r<1]+ 1)
        elif RBFtype == 'CP_C4':
            I = (r < 1)
            phi[r<1] = (1 - r(I))**6*(35/3*r[r<1]**2 + 6*r[r<1] + 1)
        """if RBFtype == 'CP_C6':
            I = (r < 1)
            phi(I) = (1 - r(I)).^8.*(32*r(I).^3 + 25*r(I).^2 + 8*r(I) + 1)
        if RBFtype == 'CTPS_C0':
            I = (r < 1)
            phi(I) = (1 - r(I)).^5
        if RBFtype == 'CTPS_C1':
            I = (r < 1 & r > 0)
            phi(I) = 1 + 80/3*r(I).^2 - 40*r(I).^3 + 15*r(I).^4 - 8/3*r(I).^5 + 20*r(I).^2.*log(r(I))
            phi(r == 0) = 1;
        if RBFtype == 'CTPS_C2a':
            I = (r < 1 & r > 0)
            phi(I) = 1 - 30*r(I).^2 - 10*r(I).^3 + 45*r(I).^4 - 6*r(I).^5 - 60*r(I).^3.*log(r(I))
            phi(r == 0) = 1
        if RBFtype == 'CTPS_C2b':
            I = (r < 1 & r > 0)
            phi(I) = 1 - 20*r(I).^2 + 80*r(I).^3 - 45*r(I).^4 -16*r(I).^5 + 60*r(I).^4.*log(r(I))
            phi(r == 0) = 1
        else:
            phi = r"""

        return phi

    def RBFeval(self, xs, x, fPar, RBFtype, R):

        Ns = len(xs[:,0]);
        dim = len(xs[0,:]);

        if len(xs[0,:]) == dim and len(fPar[0,:]) == Ns + dim + 1:
            
            N = len(x[:,0])
            
            r = np.zeros((N, Ns))
            for i in range(N):
                for j in range(Ns):
                    r[i, j] = np.norm( x[i, :] - xs[j, :] )
            
            P = [np.ones((N, 1)), x]
            M = self.radialFunction(r, RBFtype, R)
            
            y = [M, P]*fPar;
            
        else:
            
            y = []
        return y