#ifndef WTIMESOLVER_H
#define WTIMESOLVER_H

#include "vii.h"
#include "wViscFluxes.h"

namespace vii
{

class DART_API TimeSolver
{

private:
    double CFL0;
    unsigned long maxIt;
    double tol;
    unsigned int itFrozenJac0;
    bool initSln;
    double const gamma = 1.4;
    double Minf;

    ViscFluxes *SysMatrix;

public:
    TimeSolver(double _CFL0, double _Minf, double Re, unsigned long _maxIt = 100, double _tol = 1e-8, unsigned int _itFrozenJac = 5);
    ~TimeSolver();
    void InitialCondition(size_t iPoint, BLRegion *bl);
    int Integration(size_t iPoint, BLRegion *bl);

    double GetCFL0() const {return CFL0;};
    void SetCFL0(double _CFL0) {if(_CFL0>0){CFL0 = _CFL0;}};
    unsigned long GetmaxIt() const {return maxIt;};
    void SetmaxIt(unsigned long _maxIt) {maxIt = _maxIt;};
    unsigned int GetitFrozenJac() const {return itFrozenJac0;};
    void SetitFrozenJac(unsigned int _itFrozenJac) {itFrozenJac0 = _itFrozenJac;};
    void SetinitSln(bool _initSln) {initSln = _initSln;};

private:
    double SetTimeStep(double CFL, double dx, double invVel) const;
    double ComputeSoundSpeed(double gradU2) const;
    void ResetSolution(size_t iPoint, BLRegion *bl, std::vector<double> Sln0, bool usePrevPoint = false);
};
} // namespace dart
#endif //WTimeSolver_H